Require Import Arith.
Require Import ClassicalRealizability.
Require Import Integers.


Ltac Keval := basic_Keval ltac:(idtac; int_Keval fail).

(** Non deterministic boolean **)

Parameter fork : instruction.

Axiom red_fork1 : forall t u e π, red (fork↓e ★ t·u·π) (t ★ π).
Axiom red_fork2 : forall t u e π, red (fork↓e ★ t·u·π) (u ★ π).

(** We cannot put automatization tactics for evaluating fork,
    as there are two possible choices. **)
Ltac fork1 :=
  match goal with
    | |- Cst fork↓_ ★ ?t·?u·?π ∈ ⫫ => apply anti_evaluation with (t ★ π);
          [now apply red_fork1 |]
  end.

Ltac fork2 :=
  match goal with
    | |- Cst fork↓_ ★ ?t·?u·?π ∈ ⫫ => apply anti_evaluation with (u ★ π);
          [now apply red_fork2 |]
  end.

Transparent inter.

Theorem fork_spec1 : forall A B e, fork↓e ⊩ A → B → (A ∩ B).
Proof. Ksolve; [fork1 | fork2]; Ksolve. Qed.


(** Recurrence axiom **)

(* Proved with primitive integers rather than Krivine ones. *)
Definition rec := Y @ λ"n" fork @ (mℕ @ Int 0) @ ("n" @ succℕ).

Theorem recurrence_axiom : forall x e, rec↓e ⊩ ℕ x.
Proof.
Ksolve. apply (Y_realizer lt_wf). find.
Ksolve. destruct x0.
  fork1. Kevals. apply mℕ_storage. now find.
  fork2. Ksolve. now apply lt_n_Sn. start. apply succℕ_realizer.
  replace (S x0) with (x0 + 1) in Hπ. find. ring.
Qed.

(** Equivalent statements **)

Lemma equiv_fork_inter : (∀A B, A → B → (A ∩ B)) ≈ (∀A B, A → B → A) ∩ (∀A B, A → B → B).
Proof.
apply subtype_eqtype.
  apply sub_inter; subtyping.
  intros π Hπ. dec π. left. find. right. find.
Qed. 

Definition 𝔹 b : formula := ∀Z, Z 0 → Z 1 → Z b.

Lemma equiv_fork_Bool : (∀A B, A → B → (A ∩ B)) ≈ 𝔹 0 ∩ 𝔹 1.
Proof.
transitivity ((∀A B, A → B → A) ∩ (∀A B, A → B → B)). now apply equiv_fork_inter.
apply subtype_eqtype; apply inter_sub_proper.
  subtyping.
  subtyping.
  apply sub_Forall_add. intro A. apply sub_Forall_add. intro B.
  etransitivity. apply (sub_Forall _ (fun n : nat => if eq_nat_dec n 0 then A else B)). simpl. reflexivity.
  apply sub_Forall_add. intro A. apply sub_Forall_add. intro B.
  etransitivity. apply (sub_Forall _ (fun n : nat => if eq_nat_dec n 0 then A else B)). simpl. reflexivity.
Qed.

Lemma realizability_subtype : 𝔹 0 ∩ 𝔹 1 ⊆ (⊥ → ⊤ → ⊥) ∩ (⊤ → ⊥ → ⊥).
Proof.
apply inter_sub_proper.
  etransitivity. apply (sub_Forall _ (fun n : nat => if eq_nat_dec n 0 then ⊥ else ⊤)). simpl. reflexivity.
  etransitivity. apply (sub_Forall _ (fun n : nat => if eq_nat_dec n 0 then ⊤ else ⊥)). simpl. reflexivity.
Qed.

(* Equivalence ∀x, ℕ x ≈ 𝔹 0 ∩ 𝔹 1 is false with primitive integers *)
